import io.undertow.UndertowOptions;
import io.undertow.io.Sender;
import io.undertow.server.DefaultResponseListener;
import io.undertow.server.HttpHandler;
import io.undertow.server.HttpServerExchange;
import io.undertow.server.handlers.NameVirtualHostHandler;
import io.undertow.server.handlers.PathHandler;
import io.undertow.server.handlers.error.SimpleErrorPageHandler;
import io.undertow.server.protocol.http.HttpOpenListener;
import io.undertow.util.Headers;
import io.undertow.util.StatusCodes;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.xnio.BufferAllocator;
import org.xnio.ByteBufferSlicePool;
import org.xnio.ChannelListener;
import org.xnio.ChannelListeners;
import org.xnio.IoUtils;
import org.xnio.OptionMap;
import org.xnio.Options;
import org.xnio.Pool;
import org.xnio.StreamConnection;
import org.xnio.Xnio;
import org.xnio.XnioWorker;
import org.xnio.channels.AcceptingChannel;

public class ServerMain {

	private static final int MAX_BUFFER_REGION_SIZE = 10 * 1024 * 1024;

	private static OptionMap serverOptions = null;

	private static Pool<ByteBuffer> buffers = null;

	private static List<AcceptingChannel<? extends StreamConnection>> channels = Collections
			.synchronizedList(new ArrayList<AcceptingChannel<? extends StreamConnection>>());

	public static class ErrorPageHandler extends SimpleErrorPageHandler {
		@Override
		public void handleRequest(final HttpServerExchange exchange)
				throws Exception {
			exchange.addDefaultResponseListener(new DefaultResponseListener() {
				@Override
				public boolean handleDefaultResponse(
						final HttpServerExchange exchange) {
					if (!exchange.isResponseChannelAvailable()) {
						return false;
					}
					Set<Integer> codes = getResponseCodes();
					if (codes == null ? exchange.getResponseCode() >= 400
							: codes.contains(Integer.valueOf(exchange
									.getResponseCode()))) {
						final String errorPage = "<html><head><title>Error</title></head><body>Error Code: "
								+ exchange.getResponseCode()
								+ " - "
								+ StatusCodes.getReason(exchange
										.getResponseCode()) + "</body></html>";
						exchange.getResponseHeaders()
								.put(Headers.CONTENT_LENGTH,
										"" + errorPage.length());
						exchange.getResponseHeaders().put(Headers.CONTENT_TYPE,
								"text/html");
						Sender sender = exchange.getResponseSender();
						sender.send(errorPage);
						return true;
					}
					return false;
				}
			});
			getNext().handleRequest(exchange);
		}
	};

	private static ErrorPageHandler rootHandler = new ErrorPageHandler();
	private static NameVirtualHostHandler virtualHostHandler = new NameVirtualHostHandler();

	static {
		rootHandler.setNext(virtualHostHandler);

		HttpHandler notFound404Handler = new HttpHandler() {
			@Override
			public void handleRequest(HttpServerExchange exchange)
					throws Exception {
				exchange.setResponseCode(404);
			}
		};

		virtualHostHandler.setDefaultHandler(notFound404Handler);
	}

	private static PathHandler pathHandler = new PathHandler();

	static {
		HttpHandler notFound404Handler = new HttpHandler() {
			@Override
			public void handleRequest(HttpServerExchange exchange)
					throws Exception {
				exchange.setResponseCode(404);
			}
		};

		virtualHostHandler.setDefaultHandler(notFound404Handler);
	}

	/**
	 * By default the buffer size is 1k
	 */
	private static int bufferSize = 1024;

	/**
	 * Xnio instance.
	 */
	private static Xnio xnio;

	/**
	 * Xnio Worker
	 */
	private static XnioWorker worker = null;

	private static void createWorker(int ioThreadCount, int workerThreadCount) {
		if (worker == null) {
			xnio = Xnio.getInstance();
			try {
				/**
				 * Option Reference:
				 * http://www.boyunjian.com/javasrc/io.undertow
				 * /undertow-core/1.0
				 * .0.Alpha19/_/io/undertow/UndertowOptions.java
				 * https://access.redhat.com/documentation/en-US/
				 * JBoss_Enterprise_Application_Platform
				 * /6/html/API_Documentation/files/javadoc/org/xnio/Options.html
				 */
				worker = xnio
						.createWorker(OptionMap
								.builder()
								.set(Options.WORKER_IO_THREADS, ioThreadCount)
								.set(Options.WORKER_TASK_CORE_THREADS,
										workerThreadCount)
								.set(Options.WORKER_TASK_MAX_THREADS,
										workerThreadCount)
								.set(Options.TCP_NODELAY, true)
								.set(Options.REUSE_ADDRESSES, true).getMap());

			} catch (IllegalArgumentException | IOException e) {
				e.printStackTrace();
				worker = null;
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * Shutdown hook - to register with jvm for shutdown events
	 */
	private static class ShutdownHandler extends Thread {
		@Override
		public void run() {
			ServerMain.stop();
		}
	}

	private static ShutdownHandler shutdownHandler = new ShutdownHandler();

	/**
	 * Initialize the server
	 */
	public static synchronized void init() {
		Runtime.getRuntime().addShutdownHook(shutdownHandler);
	}

	/**
	 * Initialize the server
	 * 
	 * @param ioThreadCount
	 *            Input/Output thread count - These are non-blocking threads
	 * @param workerThreadCount
	 *            Worker Thread count - For any blocking operations
	 * @param buffSize
	 *            - IO buffer size. Impacts the application performance.
	 */
	public static synchronized void init(int ioThreadCount,
			int workerThreadCount, int buffSize) {

		createWorker(ioThreadCount, workerThreadCount);

		serverOptions = OptionMap.builder().set(Options.TCP_NODELAY, true)
				.set(Options.REUSE_ADDRESSES, true)
				.set(Options.WORKER_IO_THREADS, ioThreadCount).getMap();

		bufferSize = buffSize;
		buffers = new ByteBufferSlicePool(
				BufferAllocator.DIRECT_BYTE_BUFFER_ALLOCATOR, bufferSize,
				MAX_BUFFER_REGION_SIZE);

	}

	public static void addHttpListener(String host, int port)
			throws UnknownHostException, IOException {

		@SuppressWarnings("deprecation")
		HttpOpenListener openListener = new HttpOpenListener(buffers,
				OptionMap.create(UndertowOptions.BUFFER_PIPELINED_DATA, true),
				bufferSize);

		openListener.setRootHandler(rootHandler);

		virtualHostHandler.addHost(host, pathHandler);

		ChannelListener<AcceptingChannel<StreamConnection>> acceptListener = ChannelListeners
				.openListenerAdapter(openListener);
		AcceptingChannel<? extends StreamConnection> server = worker
				.createStreamConnectionServer(new InetSocketAddress(
						Inet4Address.getByName(host), port), acceptListener,
						serverOptions);

		server.resumeAccepts();

		channels.add(server);
	}

	/**
	 * Stop the server
	 */
	public static void stop() {
		if (worker == null)
			return;

		for (AcceptingChannel<? extends StreamConnection> channel : channels) {
			IoUtils.safeClose(channel);
		}
		channels = null;
		worker.shutdownNow();
		worker = null;
		xnio = null;
	}

	/**
	 * Path handler to handle request to specific path
	 * 
	 * @param path
	 *            - Path
	 * @param handler
	 *            - Handler
	 */
	@SuppressWarnings("deprecation")
	public static void addPathHandler(String path, HttpHandler handler) {
		pathHandler.addPath(path, handler);
	}

	public static void main(String args[]) {
		ServerMain.init(5, 10, 4096);
		try {
			ServerMain.addHttpListener("127.0.0.1", 8080);
			addPathHandler("/signin", new SignInHandler());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

	}
}
